'''This package contains stuff used at run-time for installing a generated
   Zope product.'''

# ------------------------------------------------------------------------------
import os, os.path, sys
import appy
import appy.version
import appy.gen as gen
from appy.gen.po import PoParser
from appy.gen.indexer import defaultIndexes, updateIndexes
from appy.gen.migrator import Migrator
from appy.gen import utils as gutils
from appy.shared.data import languages
from appy.shared.utils import getTempFileName

# ------------------------------------------------------------------------------
homePage = '<tal:h define="dummy python: request.RESPONSE.redirect(' \
           'context.config.getHomePage())"/>'
errorPage = '<tal:h define="dummy python: request.RESPONSE.write(' \
            'request.traverse(\'/\').config.getErrorPage(options))"/>'

# Zope hacks -------------------------------------------------------------------
class FakeXmlrpc:
    '''Fake class that behaves like Zope's xmlrpc module. This cheat disables
       Zope's XMLRPC.'''
    def parse_input(self, value): return None, ()
    def response(self, response): return response

class FakeZCatalog:
    # This prevents Zope frop crashing when, while updating the catalog, an
    # entry in the catalog corresponds to a missing object.
    def resolve_url(self, path, REQUEST):
        '''Cheat: "hasattr" test has been added in first line'''
        if REQUEST and hasattr(REQUEST, 'script'):
            script=REQUEST.script
            if path.find(script) != 0:
                path='%s/%s' % (script, path)
            try: return REQUEST.resolve_url(path)
            except: pass
    try:
        from Products.ZCatalog.ZCatalog import ZCatalog
        ZCatalog.resolve_url = resolve_url
    except ImportError:
        pass

class FakeSiteErrorLog:
    # Prevent Zope from logging Unauthorized errors wrapped into EvaluationError
    # instances.
    def raising(self, info):
        error = getattr(info[0], '__name__', None)
        if error and (error == 'EvaluationError'):
            orig = getattr(info[1], 'originalError', None)
            if orig.__class__.__name__ == 'Unauthorized': return
        from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
        return appy.Hack.base(SiteErrorLog.raising)
    try:
        from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
        appy.Hack.patch(SiteErrorLog.raising, raising)
    except ImportError:
        pass

def onDelSession(sessionObject, container):
    '''This function is called when a session expires'''
    rq = container.REQUEST
    if rq.cookies.has_key('_appy_') and rq.cookies.has_key('_ZopeId') and \
       (rq['_ZopeId'] == sessionObject.token):
        # The request comes from a guy whose session has expired
        resp = rq.RESPONSE
        resp.expireCookie('_appy_', path='/')
        resp.setHeader('Content-Type', 'text/html')
        resp.write('<center>For security reasons, your session has ' \
                   'expired.</center>')

# ------------------------------------------------------------------------------
class ZopeInstaller:
    '''This Zope installer runs every time Zope starts and encounters this
       generated Zope product.'''
    # Info about the default users that are always present.
    defaultUsers = {'admin': ('Manager',), 'system': ('Manager',), 'anon': ()}
    favIcon = 'favicon.ico'
    robots = 'robots.txt'

    def __init__(self, zopeContext, config, classes):
        self.zopeContext = zopeContext
        self.app = zopeContext._ProductContext__app # The root of the Zope tree
        self.config = config
        self.classes = classes
        # Unwrap some useful config variables
        self.productName = config.PROJECTNAME
        self.languages = config.appConfig.languages
        self.logger = config.logger

    def patchCssFile(self, f):
        '''p_f is an open file handler containing CSS content. We will create a
           temporary file, resolve Appy variables in it and return it.'''
        # Read the CSS content
        css = f.read()
        f.close()
        # Resolve Appy variables in it
        ui = self.config.appConfig.ui
        for name, value in ui.__dict__.iteritems():
            # Ignore non string attributes
            if name.startswith('_') or not isinstance(value, basestring):
                continue
            css = css.replace('|%s|' % name, value)
        # Dump the result in a temp file and return it
        fileName = getTempFileName(prefix='appycss', extension='css')
        f = file(fileName, 'w')
        f.write(css)
        f.close()
        f = file(fileName)
        return f

    def installUi(self):
        '''Installs the user interface'''
        # Some useful imports
        from OFS.Folder import manage_addFolder
        from OFS.Image import manage_addImage, manage_addFile
        # Delete the existing folder if it existed
        zopeContent = self.app.objectIds()
        for name in ('ui', self.favIcon, self.robots):
            if name in zopeContent: self.app.manage_delObjects([name])
        manage_addFolder(self.app, 'ui')
        # Browse the physical ui folders (the Appy one and an app-specific, if
        # the app defines one) and create the corresponding objects in the Zope
        # folder. In the case of files having the same name in both folders,
        # the one from the app-specific folder is chosen.
        j = os.path.join
        uiFolders = [j(j(appy.getPath(), 'gen'), 'ui')]
        for uiFolder in self.config.appConfig.uiFolders:
            if uiFolder.startswith('..'):
                folder = j(os.path.dirname(self.config.diskFolder),uiFolder[3:])
            else:
                folder = j(self.config.diskFolder, uiFolder)
            if os.path.exists(folder):
                uiFolders.insert(0, folder)
        for ui in uiFolders:
            for root, dirs, files in os.walk(ui):
                folderName = root[len(ui):]
                # Get the Zope folder that corresponds to this name
                zopeFolder = self.app.ui
                if folderName:
                    for name in folderName.strip(os.sep).split(os.sep):
                        zopeFolder = zopeFolder._getOb(name)
                # Create sub-folders at this level
                for name in dirs:
                    if not hasattr(zopeFolder.aq_base, name):
                        manage_addFolder(zopeFolder, name)
                # Create files at this level
                for name in files:
                    ext = os.path.splitext(name)[1]
                    if hasattr(zopeFolder.aq_base, name): continue
                    f = file(j(root, name))
                    if name == self.favIcon:
                        if not hasattr(self.app, name):
                            # Copy it at the root. Else, IE won't notice it.
                            manage_addImage(self.app, name, f)
                    elif name == self.robots:
                        # This file must be at the root
                        manage_addFile(self.app, name, f)
                    elif ext in gen.File.imageExts:
                        manage_addImage(zopeFolder, name, f)
                    elif (ext == '.css') and root.endswith('/ui'):
                        # Avoid patching CSS files from third-party products
                        manage_addFile(zopeFolder, name, self.patchCssFile(f))
                    else:
                        manage_addFile(zopeFolder, name, f)
                    f.close()
        # Update home and error pages
        from Products.PageTemplates.ZopePageTemplate import \
             manage_addPageTemplate
        if 'index_html' in zopeContent:
            self.app.manage_delObjects(['index_html'])
        manage_addPageTemplate(self.app, 'index_html', '', homePage)
        if 'standard_error_message' in zopeContent:
            self.app.manage_delObjects(['standard_error_message'])
        manage_addPageTemplate(self.app, 'standard_error_message', '',
                               errorPage)

    def installCatalog(self):
        '''Create the catalog at the root of Zope if id does not exist.'''
        if 'catalog' not in self.app.objectIds():
            # Create the catalog
            from Products.ZCatalog.ZCatalog import manage_addZCatalog
            manage_addZCatalog(self.app, 'catalog', '')
            self.logger.info('Appy catalog created.')

        # Create lexicons for ZCTextIndexes
        catalog = self.app.catalog
        lexicons = catalog.objectIds()
        from Products.ZCTextIndex.ZCTextIndex import manage_addLexicon
        if 'xhtml_lexicon' not in lexicons:
            lex = appy.Object(group='XHTML indexer', name='XHTML indexer')
            manage_addLexicon(catalog, 'xhtml_lexicon', elements=[lex])
        if 'text_lexicon' not in lexicons:
            lex = appy.Object(group='Text indexer', name='Text indexer')
            manage_addLexicon(catalog, 'text_lexicon', elements=[lex])
        if 'list_lexicon' not in lexicons:
            lex = appy.Object(group='List indexer', name='List indexer')
            manage_addLexicon(catalog, 'list_lexicon', elements=[lex])

        # Delete the deprecated one if it exists
        if 'lexicon' in lexicons: catalog.manage_delObjects(['lexicon'])

        # Create or update Appy-wide indexes and field-related indexes
        indexInfo = defaultIndexes.copy()
        tool = self.app.config
        for className in self.config.attributes.iterkeys():
            wrapperClass = tool.getAppyClass(className, wrapper=True)
            indexInfo.update(wrapperClass.getIndexes(includeDefaults=False))
        updateIndexes(self, indexInfo)
        # Re-index index "SearchableText", wrongly defined for Appy < 0.8.3
        stIndex = catalog.Indexes['SearchableText']
        if stIndex.indexSize() == 0:
            self.logger.info('reindexing SearchableText...')
            catalog.reindexIndex('SearchableText', self.app.REQUEST)
            self.logger.info('done.')

    def installBaseObjects(self):
        '''Creates the tool and the base data folder if they do not exist.'''
        # Create the tool.
        zopeContent = self.app.objectIds()
        from OFS.Folder import manage_addFolder

        if 'config' not in zopeContent:
            toolName = '%sTool' % self.productName
            gutils.createObject(self.app, 'config', toolName, self.productName,
                                wf=False, noSecurity=True)
        # Create the base data folder
        if 'data' not in zopeContent: manage_addFolder(self.app, 'data')
        # Create other base folders when relevant
        for name in self.config.appConfig.rootFolders.itervalues():
            if name not in zopeContent: manage_addFolder(self.app, name)

        # Remove some default objects created by Zope but not useful to Appy
        for name in ('standard_html_footer', 'standard_html_header',\
                     'standard_template.pt'):
            if name in zopeContent: self.app.manage_delObjects([name])

    def installTool(self):
        '''Updates the tool (now that the catalog is created) and updates its
           inner objects (users, groups, translations, documents).'''
        tool = self.app.config
        tool.createOrUpdate(True)
        appyTool = tool.appy()
        appyTool.log('Appy version is "%s".' % appy.version.short, noUser=True)

        # Execute custom pre-installation code if any
        if hasattr(appyTool, 'beforeInstall'): appyTool.beforeInstall()

        # Create the default users if they do not exist
        for login, roles in self.defaultUsers.iteritems():
            if not appyTool.count('User', noSecurity=True, login=login):
                appyTool.create('users', noSecurity=True, id=login, login=login,
                                password3=login, password4=login, roles=roles)
        # Ensure users "anon" and "system" have no email
        for login in ('anon', 'system'):
            user = appyTool.search1('User', noSecurity=True, login=login)
            user.email = None
        # Create group "admins" if it does not exist
        if not appyTool.count('Group', noSecurity=True, login='admins'):
            appyTool.create('groups', noSecurity=True, login='admins',
                            title='Administrators', roles=['Manager'])
            appyTool.log('Group "admins" created.')

        # Create a group for every global role defined in the application
        # (if required).
        if self.config.appConfig.groupsForGlobalRoles:
            for role in self.config.applicationGlobalRoles:
                groupId = role.lower()
                if appyTool.count('Group', noSecurity=True, login=groupId):
                    continue
                appyTool.create('groups', noSecurity=True, login=groupId,
                                title=role, roles=[role])
                appyTool.log('Group "%s", related to global role "%s", was ' \
                             'created.' % (groupId, role))

        # Create or update Translation objects
        translations = [t.o.id for t in appyTool.translations]
        # We browse the languages supported by this application and check
        # whether we need to create the corresponding Translation objects.
        done = []
        for language in self.languages:
            if language in translations: continue
            # We will create, in the tool, the translation object for this
            # language. Determine first its title.
            langId, langEn, langNat = languages.get(language)
            if langEn != langNat:
                title = '%s (%s)' % (langEn, langNat)
            else:
                title = langEn
            appyTool.create('translations', noSecurity=True,
                            id=language, title=title)
            done.append(language)
        if done: appyTool.log('Translations created for %s.' % ', '.join(done))

        # Synchronizes, if required, every Translation object with the
        # corresponding "po" file on disk.
        if appyTool.loadTranslationsAtStartup:
            appFolder = self.config.diskFolder
            appName = self.config.PROJECTNAME
            i18nFolder = os.path.join(appFolder, 'tr')
            done = []
            for translation in appyTool.translations:
                # Get the "po" file
                poName = '%s-%s.po' % (appName, translation.id)
                poFile = PoParser(os.path.join(i18nFolder, poName)).parse()
                for message in poFile.messages:
                    setattr(translation, message.id, message.getMessage())
                done.append(translation.id)
            appyTool.log('translation(s) %s updated (%s messages).' % \
                         (', '.join(done), len(poFile.messages)))

        # Execute custom installation code if any.
        if hasattr(appyTool, 'onInstall'): appyTool.onInstall()

    def configureSessions(self):
        '''Configure the session machinery'''
        # Register a function warning us when a session object is deleted. When
        # launching Zope in test mode, the temp folder does not exist.
        if not hasattr(self.app, 'temp_folder'): return
        sessionData = self.app.temp_folder.session_data
        if self.config.appConfig.enableSessionTimeout:
            sessionData.setDelNotificationTarget(onDelSession)
        else:
            sessionData.setDelNotificationTarget(None)

    def installZopeClasses(self):
        '''Zope-level class registration'''
        for klass in self.classes:
            name = klass.__name__
            module = klass.__module__
            wrapper = klass.wrapperClass
            exec 'from %s import manage_add%s as ctor' % (module, name)
            self.zopeContext.registerClass(meta_type=name,
                constructors = (ctor,), permission = None)
            # Create workflow prototypical instances in __instance__ attributes
            wf = wrapper.getWorkflow()
            if not hasattr(wf, '__instance__'): wf.__instance__ = wf()

    def installAppyTypes(self):
        '''We complete here the initialisation process of every Appy type of
           every gen-class of the application.'''
        appName = self.productName
        for klass in self.classes:
            # Store on wrapper class the ordered list of Appy types
            wrapperClass = klass.wrapperClass
            if not hasattr(wrapperClass, 'title'):
                # Special field "type" is mandatory for every class
                title = gen.String(multiplicity=(1,1), show='edit',
                                   indexed=True, searchable=True)
                title.init('title', None, 'appy')
                setattr(wrapperClass, 'title', title)
            # Special field "state" must be added for every class. It must be a
            # "select" field, because it will be necessary for displaying the
            # translated state name.
            state = gen.String(validator=gen.Selection('listStates'),
                           show='result', persist=False, indexed=True, height=5)
            state.init('state', None, 'workflow')
            setattr(wrapperClass, 'state', state)
            # Special field "SearchableText" must be added for every class and
            # will allow to display a search widget for entering keywords for
            # searching in index "SearchableText".
            searchable = gen.String(show=False, persist=False, indexed=True)
            searchable.init('SearchableText', None, 'appy')
            setattr(wrapperClass, 'SearchableText', searchable)
            # Set field "__fields__" on the wrapper class
            names = self.config.attributes[wrapperClass.__name__[:-8]]
            wrapperClass.__fields__ = [getattr(wrapperClass, n) for n in names]
            # Post-initialise every Appy type
            for baseClass in klass.wrapperClass.__bases__:
                if baseClass.__name__ == 'AbstractWrapper': continue
                for name, appyType in baseClass.__dict__.iteritems():
                    if not isinstance(appyType, gen.Field) or \
                           (isinstance(appyType, gen.Ref) and appyType.isBack):
                        continue # Back refs are initialised within fw refs
                    appyType.init(name, baseClass, appName)
                    if appyType.type == 'Switch':
                        # Inject switch sub-fields on the wrapper class
                        appyType.injectFields(wrapperClass)

    def installRoles(self):
        '''Installs the application-specific roles if not already done.'''
        roles = list(self.app.__ac_roles__)
        for role in self.config.applicationRoles:
            if role not in roles: roles.append(role)
        self.app.__ac_roles__ = tuple(roles)

    def patchZope(self):
        '''Patches some parts of Zope'''
        # Disable XMLRPC. This way, Zope can transmit HTTP POSTs containing
        # XML to Appy without trying to recognize it himself as XMLRPC requests.
        import ZPublisher.HTTPRequest
        ZPublisher.HTTPRequest.xmlrpc = FakeXmlrpc()

    def installDependencies(self):
        '''Zope products are installed in alphabetical order. But here, we need
           ZCTextIndex to be installed before our Appy application. So, we cheat
           and force Zope to install it now.'''
        from OFS.Application import install_product
        import Products
        install_product(self.app, Products.__path__[1], 'ZCTextIndex', [], {})

    def checkConnectedServers(self):
        '''Log the names of servers (LDAP, mail...) this app wants to connect
           to. Servers declared as "testable" are contacted to check the app's
           connection to it.'''
        cfg = self.config.appConfig
        servers = []
        tool = self.app.config.appy()
        # Browse connected servers
        for sv in ('ldap', 'mail', 'sso'):
            if not getattr(cfg, sv): continue
            svConfig = getattr(cfg, sv)
            # Initialise and check this server config
            svConfig.init(tool)
            enabled = svConfig.enabled and 'enabled' or 'disabled'
            # Test the connection to the server
            status = ''
            if svConfig.enabled and svConfig.testable:
                status = svConfig.test(tool)
                if status: status = ', ' + status
            servers.append('%s (%s, %s%s)' % (svConfig, sv, enabled, status))
        if servers:
            self.logger.info('server(s) %s configured.' % ', '.join(servers))

    def install(self):
        self.installDependencies()
        self.patchZope()
        self.installRoles()
        self.installAppyTypes()
        self.installZopeClasses()
        self.configureSessions()
        self.installBaseObjects()
        # The following line cleans and rebuilds the catalog entirely.
        #self.app.config.appy().refreshCatalog()
        self.installCatalog()
        self.installTool()
        self.installUi()
        # Check connections to external servers (ldap, mail...)
        self.checkConnectedServers()
        # Perform migrations if required
        Migrator(self).run()
        # Update Appy version in the database
        self.app.config.appy().appyVersion = appy.version.short
        # Empty the fake REQUEST object, only used at Zope startup.
        del self.app.config.getProductConfig().fakeRequest.wrappers
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>desc 2611-03</p>
<p><br /></p>
<blockquote>
<p>indentation 1</p>
<p>indentation 2</p>
<p>indentation 3</p>
</blockquote>
<p><br /></p>
<ol><li>point numéroté 1</li>
<ol><li>point numéroté 1.1</li><li>point numéroté 1.2</li></ol>
<li>point numéroté 2</li>
<ol><li>point numéroté 2.1</li><li>point numéroté 2.2</li><li>point numéroté 2.3</li>
<ol><li>point numéroté 2.3.1</li><li>point numéroté 2.3.2</li></ol>
</ol>
<li>point numéroté 3</li></ol>
<br />
<ul><li>grosse lune niveau 1</li>
<ul><li>grosse lune niveau 2</li>
<ul><li>grosse lune niveau 3</li>
<ul><li>grosse lune niveau 4<br /></li></ul>
</ul>
</ul>
</ul>
<ul>
<ul>
<ul>
<ul>
<ul><li>grosse lune niveau 5<br /></li>
<ul><li>grosse lune niveau 6<br /></li>
<ul><li>grosse lune niveau 7</li>
<ul><li>grosse lune niveau 8</li>
<ul><li>grosse lune niveau 9</li>
<ul><li>grosse lune niveau 10</li></ul>
</ul>
</ul>
</ul>
</ul>
</ul>
</ul>
</ul>
</ul>
</ul>
<br /><br />
<dl><dt>titre liste 1 </dt><dd>liste 1 </dd><dt>titre liste 2 </dt><dd>liste 2 </dd><dt>titre liste 3 </dt><dd>liste 3 </dd><dt>
<div align="center"><b>texte normal<br /></b></div>
</dt></dl>
<ol type="I"><li>romain maj 1</li><li>romain maj  2</li></ol>
<br />
<ol type="i"><li>romain 1</li><li>romain 2<br /></li></ol>
<dl>
<dl><dt><br /></dt></dl>
</dl>
<ol type="A"><li>alpha maj 1<br /></li><li>alpha maj 2</li></ol>
<br />
<ol type="a"><li>alpha min 1</li><li>alpha min 2</li></ol>
<br />blablabla<br />
<dl><dt><br /></dt></dl>
<!-- Use of CSS property 'list-style-type' -->
<ul style="list-style-type: square">
 <li>Square</li>
 <li>Square
  <ul style="list-style-type: circle">
   <li>Circle</li>
   <li>Circle</li>
  </ul>
 </li>
</ul>
<ul style="list-style-type: circle">
 <li>Circle base</li>
</ul>
'''



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os.path, io
from base64 import encodebytes
from xml.sax.saxutils import quoteattr

from appy.xml import xmlPrologue
from appy.xml.escape import Escape

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Marshaller:
    '''Allows to produce the XML version of a Python object. The XML result
       respects conventions as described in appy.xml.unmarshaller.py.'''

    # Names of object attributes to exclude from the result, by class name
    fieldsToExclude = {'History': 'o'}

    # The correspondence between classes and their equivalent base class
    typesMap = {'list': 'list', 'PersistentList': 'list', 'UserList': 'list',
                'dict': 'dict', 'PersistentMapping': 'dict', 'UserDict': 'dict',
                'FileInfo': 'file', 'bool': 'bool', 'int': 'int',
                'float': 'float', 'bytes': 'bytes', 'tuple': 'tuple',
                'DateTime': 'DateTime'}

    def __init__(self, cdata=False, conversionFunctions={}, rootTag='appyData',
                 dumpXmlPrologue=True, namespaces={}, namespacedTags={}):
        # If p_cdata is True, all string values will be dumped as XML CDATA
        self.cdata = cdata
        # The following dict stores specific conversion (=Python to XML)
        # functions. Such functions are useful when you are not happy with the
        # way built-in converters work, or if you want to define a specific way
        # to represent, in XML, some particular Python object or value. In this
        # dict, every key represents a given type (it must correspond to the
        # class name as retrieved by someObject.__class__.__name__); every value
        # is a function accepting 2 args: the first one is the StringIO where
        # the result is being dumped, while the second one is the Python object
        # to dump.
        self.conversionFunctions = conversionFunctions
        # If dumpXmlPrologue is True, the XML prologue will be dumped
        self.dumpXmlPrologue = dumpXmlPrologue
        # The name of the root tag
        self.rootTagName = rootTag
        # The namespaces that will be defined at the root of the XML message.
        # It is a dict whose keys are namespace prefixes and whose values are
        # namespace URLs. If you want to specify a default namespace, specify an
        # entry whose key is an empty string.
        self.namespaces = namespaces
        # The following dict will tell which XML tags will get which namespace
        # prefix ({s_tagName: s_prefix}). Special optional dict entry
        # '*':s_prefix will indicate a default prefix that will be applied to
        # any tag that does not have it own key in this dict.
        self.namespacedTags = namespacedTags
        # The following attribute will hold the object to marshall (will be set
        # by m_parshall).
        self.o = None

    def getTagName(self, name):
        '''Returns the name of tag p_name as will be dumped. It can be p_name,
           or p_name prefixed with a namespace prefix (will depend on
           self.prefixedTags).'''
        # Determine the prefix
        prefix = ''
        if name in self.namespacedTags: prefix = self.namespacedTags[name]
        elif '*' in self.namespacedTags: prefix = self.namespacedTags['*']
        return '%s:%s' % (prefix, name) if prefix else name

    def isAppy(self, o):
        '''Returns True if p_o is an Appy object'''
        return hasattr(o, 'class_')

    def isObject(self, o):
        '''Returns True if p_o is an instance of a custom class, False if it is
           a basic type, or tuple, sequence, etc.'''
        return hasattr(o, '__dict__')

    def dumpRootTag(self, r, o):
        '''Dumps the root tag'''
        # Dumps the name of the tag
        tagName = self.getTagName(self.rootTagName)
        r.write('<%s type="object"' % tagName)
        # Dumps namespace definitions if any
        for prefix, url in self.namespaces.items():
            if not prefix:
                pre = 'xmlns' # The default namespace
            else:
                pre = 'xmlns:%s' % prefix
            r.write(' %s="%s"' % (pre, url))
        # Dumps Appy-specific attributes
        if self.isAppy(o):
            r.write(' id="%s" iid="%d" className="%s"' % \
                    (o.id, o.iid, o.class_.name))
        r.write('>')
        return tagName

    def dumpString(self, r, s):
        '''Dumps a string into the result'''
        # Surround it with a CDATA when appropriate
        if self.cdata: r.write('<![CDATA[')
        # Escape XML chars
        r.write(Escape.xml(s))
        if self.cdata: r.write(']]>')

    def dumpFile(self, r, v):
        '''Dumps a file into the result'''
        if not v: return
        w = r.write
        # p_value contains an instance of class appy.model.fields.file.FileInfo.
        # Encode it in Base64, in one or several parts.
        partTag = self.getTagName('part')
        r.write('<%s type="base64" number="1">' % partTag)
        if v.inDb():
            path = v.getFilePath(self.o)
        else:
            path = v.fsPath
        if not os.path.isfile(path):
            w('</%s>' % partTag) # Close the (empty) tag
            return
        f = open(path, 'rb')
        partNb = 1
        while True:
            chunk = f.read(v.BYTES)
            if not chunk: break
            # We have one more chunk. Dump the start tag (excepted if it is
            # the first chunk: the start tag has already been dumped, see
            # above).
            if partNb > 1:
                w('<%s type="base64" number="%d">' % (partTag, partNb))
            w(encodebytes(chunk).decode())
            w('</%s>' % partTag) # Close the tag
            partNb += 1
        f.close()

    def dumpDict(self, r, value):
        '''Dumps in p_r the XML version of dict p_value'''
        for k, v in value.items():
            r.write('<entry type="object">')
            self.dumpField(r, 'k', k)
            self.dumpField(r, 'v', v)
            r.write('</entry>')

    def dumpList(self, r, value):
        '''Dumps the XML version of list p_value'''
        for v in value: self.dumpField(r, 'e', v)

    # Dumping a tuple is similar to dumping a list
    dumpTuple = dumpList

    def dumpBool(self, r, value):
        '''Dumps in p_r this p_boolean value'''
        r.write(str(bool(value)))

    def dumpBytes(self, r, value):
        '''Dumps p_value, which is a bytes value'''
        r.write(value.decode())

    def dumpValue(self, r, value, type, className):
        '''Dumps in p_r the XML version of p_value'''
        # Use a custom function if one is defined for this type of value
        if className in self.conversionFunctions:
            self.conversionFunctions[className](r, value)
            return
        # Find the specific method for dumping this p_value
        if not type:
            r.write(value or '')
        else:
            method = 'dump%s' % type.capitalize()
            if hasattr(self, method):
                exec('self.%s(r, value)' % method)
            else:
                r.write(str(value))

    def dumpField(self, r, name, value):
        '''Dumps in p_r, the p_value of field named p_name'''
        # As a preamble, manage special case of p_name being "_any". In that
        # case, p_value corresponds to a previously marshalled string that must
        # be included as is here, without dumping the tag name.
        if name == '_any':
            r.write(value)
            return
        # Dump the start tag
        tagName = self.getTagName(name)
        r.write('<%s' % tagName)
        # Dump value's type as an XML attribute
        className = value.__class__.__name__
        if className in self.typesMap:
            type = self.typesMap[className]
        elif self.isObject(value):
            type = 'object'
        else:
            # For any other type, attribute "type" will not be dumped
            type = None
        if type: r.write(' type="%s"' % type)
        # Dump the value's length if multi-valued
        if type in ('list', 'tuple'):
            length = len(value) if value else 0
            r.write(' count="%d"' % length)
        # Dump file-related attributes
        if value and (type == 'file'):
            # Dump the MIME type
            r.write(' mimeType="%s"' % value.mimeType)
            # Dump the file name
            r.write(' name=%s' % quoteattr(value.uploadName))
        r.write('>')
        # Dump the field value
        self.dumpValue(r, value, type, className)
        # Dump the end tag
        r.write('</%s>' % tagName)

    def mustDump(self, name, className):
        '''Must attribute named p_name on the class named p_className be part of
           the result ?'''
        exclude = self.fieldsToExclude.get(className)
        return not exclude or (name not in exclude)

    def dumpObject(self, r, o, complete=False, fieldNames=None):
        '''Dumps this p_o(bject) in p_r'''
        # Is p_o an Appy object ?
        # ----------------------------------------------------------------------
        #  yes | if p_complete is True, the entire object is dumped, with all
        #      | its fields being visible on the "xml" layout (p_fieldNames is
        #      | ignored). If p_complete is False, only its URL is dumped.
        # ----------------------------------------------------------------------
        #  no  | p_complete is ignored and the entire object is dumped, based on
        #      | its __dict__ or using names from p_fieldNames.
        # ----------------------------------------------------------------------
        if self.isAppy(o):
            # It is a Appy object
            if complete:
                className = o.class_.name
                # Browse p_o's fields that must appear on the XML layout
                for field in o.getFields('xml'):
                    # Dump only needed fields
                    if not self.mustDump(field.name, className): continue
                    val = field.getValue(o, single=False)
                    v = field.getXmlValue(o, val)
                    self.dumpField(r, field.name, v)
                # Dump the object history
                self.dumpField(r, 'history', o.history)
            else:
                # Dump its URL
                r.write('%s/xml' % o.url)
        else:
            # A non-Appy object: dump it in its entirety
            className = o.__class__.__name__
            if fieldNames:
                # The fields are specified in a list, in order to dump them in
                # that specific order.
                for name in fieldNames:
                    if self.mustDump(name, className):
                        value = getattr(o, name, None)
                        self.dumpField(r, name, value)
            else:
                # Get the field by browsing p_o's dict
                for name, value in o.__dict__.items():
                    if self.mustDump(name, className):
                        self.dumpField(r, name, value)

    def marshall(self, o, appy=False, conversionFunctions=None,
                 fieldNames=None):
        '''Returns, as a string, the XML version of this p_o(bject)'''
        self.o = o
        if conversionFunctions:
            self.conversionFunctions.update(conversionFunctions)
        # Create the buffer where the XML result will be dumped
        r = io.StringIO()
        # Dump the XML prologue if required
        if self.dumpXmlPrologue:
            r.write(xmlPrologue)
        if self.isObject(o):
            # Dump the root tag
            rootTagName = self.dumpRootTag(r, o)
            # Dump the fields of this root object
            self.dumpObject(r, o, complete=True, fieldNames=fieldNames)
            self.marshallSpecificElements(o, r)
            r.write('</%s>' % rootTagName)
        else:
            self.dumpField(r, self.rootTagName, o)
        # Return the result
        return r.getvalue()

    def marshallSpecificElements(self, o, r):
        '''You can use this marshaller as a base class for creating your own.
           In this case, this method will be called by m_marshall to allow your
           concrete marshaller to insert more things in the result. p_r is the
           StringIO buffer where the result of the marshalling process is
           currently dumped; p_o is the object being currently marshalled.'''
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>Some HTML entities: é: &eacute;, è: &egrave;, Atilde: &Atilde;.</p>
<p>XML entities: amp: &amp;, quote: &quot;, apos: &apos;, lt: &lt;, gt: &gt;.</p>
<p>&nbsp;</p><p>Para</p>'''



'''This sript allows to wrap a Python module into an egg.'''

# ------------------------------------------------------------------------------
import os, os.path, sys, zipfile, appy
from appy.bin.clean import Cleaner
from appy.shared.utils import FolderDeleter, copyFolder, cleanFolder
from optparse import OptionParser

# ------------------------------------------------------------------------------
class EggifierError(Exception): pass
ERROR_CODE = 1
eggInfo = '''from setuptools import setup, find_packages
import os
setup(name = "%s", version = "%s", description = "%s",
      long_description = "%s",
      author = "%s", author_email = "%s",
      license = "GPL", keywords = "plone, appy", url = '%s',
      classifiers = ["Framework :: Appy", "Programming Language :: Python",],
      packages=find_packages(exclude=['ez_setup']), include_package_data = True,
      namespace_packages=['%s'], zip_safe = False,
      install_requires=['setuptools'],)'''
initInfo = '''
# See http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)
'''

# ------------------------------------------------------------------------------
class EggifyScript:
    '''usage: python eggify.py pythonModule [options]
       pythonModule is the path to a Python module or the name of a Python file.

       Available options are:
         -a --appy      If specified, the Appy module (light version, without
                        test code) will be included in the egg.
         -r --result    The path where to create the egg (defaults to the
                        current working directory)
         -p --products  If specified, the module will be packaged in the
                        "Products" namespace.
         -v --version   Egg version. Defaults to 1.0.0.
    '''
    def createSetupFile(self, eggTempFolder):
        '''Creates the setup.py file in the egg.'''
        content = eggInfo % (self.moduleName, self.version, 'Appy module',
                             'Appy module', 'Gaetan Delannay',
                             'gaetan.delannay AT gmail.com',
                             'http://appyframework.org',
                             self.moduleName.split('.')[0])
        f = file(os.path.join(eggTempFolder, 'setup.py'), 'w')
        f.write(content)
        f.close()

    def createInitFile(self, eggTempFolder):
        '''Creates the ez_setup-compliant __init__ files.'''
        initPath = os.path.join(eggTempFolder,self.moduleName.split('.')[0])
        f = file(os.path.join(initPath, '__init__.py'), 'w')
        f.write(initInfo)
        f.close()

    def getEggName(self):
        '''Creates the egg name.'''
        return '%s-%s.egg' % (self.moduleName, self.version)

    zipExclusions = ('.svn', 'doc', 'test', 'versions')
    def dirInZip(self, dir):
        '''Returns True if the p_dir must be included in the zip.'''
        for exclusion in self.zipExclusions:
            if dir.endswith(exclusion) or ('/%s/' % exclusion in dir):
                return False
        return True
    
    def zipResult(self, eggFullName, eggTempFolder):
        '''Zips the result and removes the egg temp folder.'''
        zipFile = zipfile.ZipFile(eggFullName, 'w', zipfile.ZIP_DEFLATED)
        # Put the Python module inside the egg.
        prefix = os.path.dirname(eggTempFolder)
        for dir, dirnames, filenames in os.walk(eggTempFolder):
            for f in filenames:
                fileName = os.path.join(dir, f)
                zipFile.write(fileName, fileName[len(prefix):])
        # Put the Appy module inside it if required.
        if self.includeAppy:
            eggPrefix = '%s/%s' % (eggTempFolder[len(prefix):],
                                   self.moduleName.replace('.', '/'))
            # Where is Appy?
            appyPath = os.path.dirname(appy.__file__)
            appyPrefix = os.path.dirname(appyPath)
            # Clean the Appy folder
            Cleaner().run(verbose=False)
            # Insert appy files into the zip
            for dir, dirnames, filenames in os.walk(appyPath):
                if not self.dirInZip(dir): continue
                for f in filenames:
                    fileName = os.path.join(dir, f)
                    zipName = eggPrefix + fileName[len(appyPrefix):]
                    zipFile.write(fileName, zipName)
        zipFile.close()
        # Remove the temp egg folder.
        FolderDeleter.delete(eggTempFolder)

    def eggify(self):
        '''Let's wrap a nice Python module into an ugly egg.'''
        j = os.path.join
        # First, clean the Python module
        cleanFolder(self.pythonModule, verbose=False)
        # Create the egg folder
        eggFullName = j(self.eggFolder, self.eggName)
        if os.path.exists(eggFullName):
            os.remove(eggFullName)
            print('Existing "%s" was removed.' % eggFullName)
        # Create a temp folder where to store the egg
        eggTempFolder = os.path.splitext(eggFullName)[0]
        if os.path.exists(eggTempFolder):
            FolderDeleter.delete(eggTempFolder)
            print('Removed "%s" that was in my way.' % eggTempFolder)
        os.mkdir(eggTempFolder)
        # Create the "Products" sub-folder if we must wrap the package in this
        # namespace
        eggModulePath = j(j(eggTempFolder, self.moduleName.replace('.', '/')))
        # Copy the Python module into the egg.
        os.makedirs(eggModulePath)
        copyFolder(self.pythonModule, eggModulePath)
        # Create setup files in the root egg folder
        self.createSetupFile(eggTempFolder)
        self.createInitFile(eggTempFolder)
        self.zipResult(eggFullName, eggTempFolder)

    def checkArgs(self, options, args):
        # Check that we have the correct number of args.
        if len(args) != 1: raise EggifierError('Wrong number of arguments.')
        # Check that the arg corresponds to an existing Python module
        if not os.path.exists(args[0]):
            raise EggifierError('Path "%s" does not correspond to an ' \
                                'existing Python package.' % args[0])
        self.pythonModule = args[0]
        # At present I only manage Python modules, not 1-file Python packages.
        if not os.path.isdir(self.pythonModule):
            raise EggifierError('"%s" is not a folder. One-file Python ' \
                                'packages are not supported yet.' % args[0])
        self.eggFolder = options.result
        if not os.path.exists(self.eggFolder):
            raise EggifierError('"%s" does not exist. Please create this ' \
                                'folder first.' % self.eggFolder)
        self.includeAppy = options.appy
        self.inProducts = options.products
        self.version = options.version
        self.moduleName = os.path.basename(self.pythonModule)
        if self.inProducts:
            self.moduleName = 'Products.' + self.moduleName
        self.eggName = self.getEggName()

    def run(self):
        optParser = OptionParser(usage=EggifyScript.__doc__)
        optParser.add_option("-r", "--result", dest="result",
                             help="The folder where to create the egg",
                             default=os.getcwd(), metavar="RESULT",
                             type='string')
        optParser.add_option("-a", "--appy", action="store_true",
                             help="Includes the Appy module in the egg")
        optParser.add_option("-p", "--products", action="store_true",
                             help="Includes the module in the 'Products' " \
                                  "namespace")
        optParser.add_option("-v", "--version", dest="version",
                             help="The module version", default='1.0.0',
                             metavar="VERSION", type='string')
        options, args = optParser.parse_args()
        try:
            self.checkArgs(options, args)
            self.eggify()
        except EggifierError, ee:
            sys.stderr.write(str(ee) + '\nRun eggify.py -h for getting help.\n')
            sys.exit(ERROR_CODE)

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    EggifyScript().run()
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import copy

from appy import Object
from appy.px import Px
from appy.fields.list import List
from appy.gen.layout import Table
from appy.shared.diff import HtmlDiff

# ------------------------------------------------------------------------------
class SepRow:
    '''Represents a custom row that is a separator but does not represent
       data.'''
    def __init__(self, text, cellCss=None, rowCss=None):
        self.text = text
        # An optional CSS class to apply to the cell containing p_text
        self.cellCss = cellCss
        # An optional CSS class to apply to the row contaning this cell
        self.rowCss = rowCss

    def get(self, field):
        '''Produces the chunk of XHTML code representing this row'''
        rowCss = self.rowCss and (' class="%s"' % self.rowCss) or ''
        cellCss = self.cellCss and (' class="%s"' % self.cellCss) or ''
        return '<tr valign="top"%s><td colspan="%s"%s>%s</td></tr>' % \
               (rowCss, len(field.fields) + 1, cellCss, self.text)

# ------------------------------------------------------------------------------
class Dict(List):
    '''A Dict value has the form ~{s_key: Object}~. Keys are fixed and given by
       a method specified in parameter "keys". Values are Object instances,
       whose attributes are determined by parameter "fields" that, similarly to
       the List field, determines sub-data for every entry in the dict. This
       field is build on top of the List field.'''
    SepRow = SepRow

    # PX for rendering a single row
    pxRow = Px('''
     <!-- Render a separatation row -->
     <x if="not rowId">::text.get(field)</x>
     <tr if="rowId" valign="top" class=":loop.rowId.odd and 'even' or 'odd'">
      <x>:field.pxFirstCell</x>
      <td class="discreet">::text</td>
      <td for="subName, field in subFields" if="field" align="center"
          var2="fieldName='%s*%s' % (field.name, rowId)">:field.pxRender</td>
     </tr>''')

    # PX for rendering the dict (shared between pxView and pxEdit)
    pxTable = Px('''
     <table var="isEdit=layoutType == 'edit'" if="isEdit or value"
            id=":'list_%s' % name" class="grid" width=":field.width"
            var2="keys=field.keys(obj);
                  subFields=field.getSubFields(zobj, layoutType);
                  zobj=oobj|zobj">
      <!-- Header -->
      <tr valign="bottom">
       <th width=":field.widths[0]"></th>
       <th for="subName, sub in subFields" if="sub"
           width=":field.widths[loop.subName.nb + 1]">::_(sub.labelId)</th>
      </tr>
      <!-- Rows of data -->
      <x for="rowId, text in keys">:field.pxRow</x>
     </table>''')

    def __init__(self, keys, fields, validator=None, multiplicity=(0,1),
      default=None, defaultOnEdit=None, show=True, page='main', group=None,
      layouts=None, move=0, specificReadPermission=False,
      specificWritePermission=False, width='', height=None, maxChars=None,
      colspan=1, master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None,
      subLayouts=Table('frv', width=None), widths=None, view=None, cell=None,
      edit=None, xml=None, translations=None):
        List.__init__(self, fields, validator, multiplicity, default,
          defaultOnEdit, show, page, group, layouts, move,
          specificReadPermission, specificWritePermission, width, height,
          maxChars, colspan, master, masterValue, focus, historized, mapping,
          generateLabel, label, subLayouts, widths, view, cell, edit, xml,
          translations)
        # Method in "keys" must return a list of tuples (key, title): "key"
        # determines the key that will be used to store the entry in the
        # database, while "title" will get the text that will be shown in the ui
        # while encoding/viewing this entry.

        # WARNING: a key must be a string, cannot contain digits only and
        # cannot contain char "*". A key is typically an object ID.

        # For a nice rendering of your dict, some of the tuples returned by
        # method "keys" can be "separator rows". The tuple representing such a
        # row must have the form (None, sepRow). "None" indicates that this is
        # not a row of data; "sepRow" must be a SepRow instance (see hereabove)
        # that will determine content and style for the separator row.
        self.keys = keys

    def computeWidths(self, widths):
        '''Set given p_widths or compute default ones if not given'''
        if not widths:
            self.widths = [''] * (len(self.fields) + 1)
        else:
            self.widths = widths

    def getStorableValue(self, obj, value, complete=False):
        '''Gets p_value in a form that can be stored in the database'''
        from persistent.mapping import PersistentMapping
        res = PersistentMapping()
        for k, v in value.iteritems():
            res[k] = self.getStorableRowValue(obj, v)
        return res

    def getComparableValue(self, obj):
        '''Return a (deep) copy of field value on p_obj'''
        # Indeed, because a new value does not overwrite but updates the stored
        # value, the comparable value must be a deep copy of the stored value.
        r = getattr(obj.aq_base, self.name, None)
        if r: return copy.deepcopy(r)

    def remove(self, obj, key):
        '''Remove entry corresponding to p_key on the value stored on p_obj'''
        val = getattr(obj.o, self.name, None)
        if not val: return
        if key not in val: return
        del val[key]
        setattr(obj.o, self.name, val)

    def store(self, obj, value, overwrite=False):
        '''Stores the p_value (produced by m_getStorableValue) on p_obj. If some
           entry from p_value already exists in the DB value, it is updated,
           not overwritten.'''
        obj = obj.o
        if not self.persist: return
        dbValue = getattr(obj.aq_base, self.name, None)
        if not dbValue or overwrite:
            setattr(obj, self.name, value)
        else:
            # Update the DB value with p_value
            if not value: return
            for key, data in value.iteritems():
                if key not in dbValue:
                    dbValue[key] = data
                else:
                    dbValue[key].update(data)
                    # Force the mapping to take the change into account
                    dbValue[key] = dbValue[key]
            setattr(obj, self.name, dbValue)

    def subValidate(self, obj, value, errors):
        '''Validates inner fields'''
        for key, row in value.iteritems():
            for name, subField in self.fields:
                message = subField.validate(obj, getattr(row, name, None))
                if message:
                    setattr(errors, '%s*%s' % (subField.name, key), message)

    def getDiffSubValue(self, old, new, texts):
        '''Produce a diff between this p_old sub-value and its p_new version.
           p_texts contain precomputed translated texts about the user that
           performed the change.'''
        if old:
            r = HtmlDiff.templateDiv % (HtmlDiff.deleteStyle, texts[1], old)
        else:
            r = ''
        if new:
            r = '%s%s' % (r, HtmlDiff.templateDiv % (HtmlDiff.insertStyle,
                                                     texts[0], new))
        return r

    def getDiffValue(self, obj, old, new, texts):
        '''Return a XHTML representation of dict value p_new incorporating the
           differences with this p_old version.'''
        # Return the p_old value as-is, if the p_new version is None
        if not new:
            fake = self.getFakeObject(old, obj.REQUEST)
            return self.doRender('view', fake)
        # Update the old value by replacing every modified sub-value (w.r.t
        # p_new) with a diff between the old and the new sub-value.
        diff = copy.deepcopy(old)
        for key, sub in old.iteritems():
            # Get the corresponding entry in p_new
            newSub = new.get(key)
            diffSub = diff[key]
            if newSub:
                # Browse sub-values on this row of data
                for name, val in sub.d().iteritems():
                    # Get the new version of sub-value
                    newVal = getattr(newSub, name, None)
                    if newVal == val: continue
                    # Values are different, produce a diff
                    diffVal = self.getDiffSubValue(val, newVal, texts)
                    setattr(diffSub, name, diffVal)
            else:
                # This complete row has been removed from p_new
                for name, val in sub.d().iteritems():
                    diffVal = self.getDiffSubValue(val, None, texts)
                    setattr(diffSub, name, diffVal)
        # Produce a fake object containing the diff
        fake = self.getFakeObject(diff, obj.REQUEST)
        # Render the diff
        return self.doRender('view', obj.appy(), minimal=True,
                             specific={'oobj':fake})
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from DateTime import DateTime

from appy.px import Px
from appy.model.fields import Field

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
INVALID_MULTILINGUAL_VALUE = 'Multilingual field "%s" accepts a dict whose ' \
  'keys are in field.languages and whose values are strings.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Multilingual:
    '''Mixin class injected into any Field whose content can be multilingual'''

    # Default values to render when field values are empty
    emptyDefault = {'view': '-', 'edit': ''}
    # Default top space (in pixels) to apply in pxLanguage
    lgTop = {'view': 1, 'edit': 3}

    # Note that multilinguality is a dynamic feature: field values can be
    # unilingual or multilingual, depending on some condition on the container
    # object itself, or on some site-specific configuration.

    # PX displaying the language code and name besides the part of the
    # multilingual field storing content in this language.
    pxLanguage = Px('''
     <td style=":'padding-top:%dpx' % field.lgTop[layout]" width="25px">
      <span class="language help"
            title=":ui.Language.getName(lg)">:lg.upper()</span>
     </td>''')

    # PX acting as a substitute for the field pxView. This PX determines if the
    # field content is multilingual or not. If the field is unilingual, it
    # simply calls a PX named "<layoutType>Uni" on the field. Else, it renders
    # such a PX for every supported language, calling "<layoutType>Uni" for
    # every language, and assembling the result according to the "languages
    # layout". The "Uni" PX receives the current language as variable "lg". If
    # the field is unilingual, the received "lg" variable is None.
    view = edit = cell = Px('''
     <x var="languages=field.getAttribute(o, 'languages');
             multilingual=len(languages) &gt; 1;
             pxUni=getattr(field, '%sUni' % layout)">

      <!-- Display the uni-lingual version of the field -->
      <x if="not multilingual" var2="lg=None">:pxUni</x>

      <!-- Display the multi-lingual version of the field -->
      <x if="multilingual"
         var2="languageLayout=field.getLanguagesLayout(layout)">

       <!-- Horizontally-layouted multilingual field -->
       <table if="languageLayout == 'horizontal'" width="100%"
              class=":(layout == 'cell') and 'no' or ''"
              var="count=len(languages)">
        <tr valign="top"><x for="lg in languages"><x>:field.pxLanguage</x>
         <td width=":'%d%%' % int(100.0/count)"
             var="requestValue=requestValue[lg]|None;
                  value=value[lg]|field.emptyDefault[layout]">:pxUni</td>
        </x></tr></table>

       <!-- Vertically-layouted multilingual field -->
       <table if="languageLayout == 'vertical'"
              class=":(layout == 'cell') and 'no' or ''">
        <tr valign="top" height="20px" for="lg in languages">
         <x>:field.pxLanguage</x>
         <td var="requestValue=requestValue[lg]|None;
                  value=value[lg]|field.emptyDefault[layout]">:pxUni</td>
       </tr></table>
      </x>
     </x>''')

    def __init__(self, languages, languagesLayouts):
        '''Inject multilingual-specific attributes on p_self'''
        # If "languages" holds more than one language, the field will be
        # multi-lingual and several widgets will allow to edit/visualize the
        # field content in all the supported languages. The field is also used
        # by the CK spell checker.
        self.languages = languages
        # When content exists in several languages, how to render them? Either
        # horizontally (one below the other), or vertically (one besides the
        # other). Specify here a dict whose keys are layouts ("edit", "view")
        # and whose values are either "horizontal" or "vertical".
        self.languagesLayouts = languagesLayouts

    def isMultilingual(self, o, dontKnow=False):
        '''Is this field multilingual ? If we don't know, say p_dontKnow'''
        if not o:
            if callable(self.languages):
                # In that case, it is impossible to know
                return dontKnow
            else: return len(self.languages) > 1
        return len(self.getAttribute(o, 'languages')) > 1

    def getLanguagesLayout(self, layout):
        '''Gets the way to render a multilingual field on p_layoutType.'''
        if self.languagesLayouts and (layout in self.languagesLayouts):
            return self.languagesLayouts[layout]
        # Else, return a default value that depends of the format
        return self.defaultLanguagesLayouts[layout]

    def getCopyValue(self, o):
        '''A value being multilingual is stored in a dict. For such a value,
           standard method Field.getCopyValue must return a distinct copy of the
           value as stored on p_obj.'''
        r = self.getValue(o)
        if isinstance(r, dict): r = r.copy()
        return r

    def updateHistoryValue(self, o, values):
        '''If p_self is multilingual, remove its dict value from p_values and
           set, instead, a key for every language-specific value.'''
        languages = self.getAttribute(o, 'languages')
        if len(languages) > 1:
            # Consider every language-specific value as a first-class value:
            # delete the multi-language dict value and replace it with distinct
            # language-specific values, with a special key including the
            # language.
            value = values[self.name]
            del(values[self.name])
            for lg in languages:
                lgPrev = value.get(lg)
                lgCurr = o.values.get(self.name).get(lg)
                if lgPrev != lgCurr:
                    values[(self.name, lg)] = lgPrev

    def valueIsInRequest(self, o, req, name=None, layout='view'):
        '''Multilingual values are stored in specific input fields with
           specific names.'''
        # If we are on the search layout, p_obj, if not None, is certainly not
        # the p_obj we want here (can be a home object).
        if layout == 'search':
            return Field.valueIsInRequest(self, o, req, name, layout)
        languages = self.getAttribute(o, 'languages')
        if len(languages) == 1:
            return Field.valueIsInRequest(self, o, req, name, layout)
        # Is is sufficient to check that at least one of the language-specific
        # values is in the request.
        return ('%s_%s' % (name, languages[0])) in req

    def getRequestValue(self, o, requestName=None):
        '''The request value may be multilingual'''
        req = o.req
        name = requestName or self.name
        languages = self.getAttribute(o, 'languages')
        # A unilingual field
        if len(languages) == 1: return req[name]
        # A multilingual field
        r = {}
        for language in languages:
            r[language] = req['%s_%s' % (name, language)]
        return r

    def isEmptyValue(self, o, value):
        '''Returns True if the p_value must be considered as an empty value'''
        if not isinstance(value, dict):
            return Field.isEmptyValue(self, o, value)
        # p_value is a dict of multilingual values. For such values, as soon
        # as a value is not empty for a given language, the whole value is
        # considered as not being empty.
        for v in value.values():
            if not Field.isEmptyValue(self, o, v): return
        return True

    def isCompleteValue(self, o, value):
        '''Returns True if the p_value must be considered as complete. For a
           unilingual field, being complete simply means not being empty. For a
           multilingual field, being complete means that a value is present for
           every language.'''
        if not self.isMultilingual(o):
            return Field.isCompleteValue(self, o, value)
        # As soon as a given language value is empty, the global value is not
        # complete. If p_value is a string, we are validating a sub-value for a
        # specific language.
        if not value or isinstance(value, str): return True
        for v in value.values():
            if Field.isEmptyValue(self, o, v): return
        return True

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        '''The multilingual and unilingual variants of p_value's formatted
           version differ.'''
        # Note that p_language represents the UI language, while variable
        # "languages" below represents the content language(s) of this field.
        languages = self.getAttribute(o, 'languages')
        uni = self.getUniFormattedValue
        if (len(languages) == 1) or isinstance(value, str):
            # Normally, p_value should not be a string if there is a single
            # language. This can happen in exceptional cases, ie, in a
            # object's history (data change), when an object was transmitted
            # from one App1 to App2, where a field is unilingual in App1 and
            # multilingual in App2.
            return uni(o, value, layout, showChanges, language=language)
        # Return the dict of values whose individual, language-specific values
        # have been formatted via m_getUniFormattedValue.
        if not value and not showChanges: return value
        r = {}
        for lg in languages:
            if not value: val = ''
            else: val = value[lg]
            r[lg] = uni(o, val, layout, showChanges,
                        language=language, contentLanguage=lg)
        return r

    def getShownValue(self, o, value, layout='view', showChanges=False,
                      language=None):
        '''For a multilingual field, this method only shows one specific
           language part.'''
        # Be careful: p_language represents the UI language, while variable
        # "languages" below represents the content language(s) of this field.
        languages = self.getAttribute(o, 'languages')
        uni = self.getUniFormattedValue
        if len(languages) == 1:
            return uni(o, value, layout, showChanges, language=language)
        if not value: return value
        # Try to propose the part that is in the user language, or the part of
        # the first content language else.
        lg = o.guard.userLanguage
        if lg not in value: lg = languages[0]
        return uni(o, value[lg], layout, showChanges, language=lg)

    def getIndexValue(self, o, searchable=False, raw=False):
        '''Multilingual variant of Field::getIndexValue. Language parts must be
           concatenated into a single value before being indexed.'''
        r = Field.getIndexValue(self, o, searchable=searchable, raw=True)
        if not r: return
        # Manage multilinguality
        if isinstance(r, dict):
            vals = []
            for val in r.values():
                if val:
                    vals.append(val)
            r = ' '.join(vals)
        # Get the value corresponding to the internal index representation
        index = self.getIndexType(True)
        return index.toString(o, r) if searchable else index.toIndexed(r, self)

    def getStorableValue(self, o, value, complete=False):
        # If no object is passed, assume the field is unilingual
        if o is None:
            return self.getUniStorableValue(o, value)
        # Check unilinguality
        languages = self.getAttribute(o, 'languages')
        if len(languages) == 1:
            return self.getUniStorableValue(o, value)
        # A multilingual value is stored as a dict whose keys are ISO 2-letters
        # language codes and whose values are strings storing content in the
        # language ~{s_language: s_content}~.
        if not value: return
        # p_value may hold a part of the complete value, in some language
        if isinstance(value, str):
            return self.getUniStorableValue(o, value)
        # Manage a multilingual p_value
        for lg in languages:
            value[lg] = self.getUniStorableValue(o, value[lg])
        return value

    def validateValue(self, o, value):
        '''Validates this p_value'''
        if isinstance(value, dict):
            r = []
            for lang, val in value.items():
                res = self.validateUniValue(o, val)
                if res:
                    r.append('%s (%s)' % (res, lang.upper()))
            r = '\n'.join(r) if r else None
        else:
            r = self.validateUniValue(o, value)
        return r

    def store(self, o, value):
        '''Stores p_value on p_o for this field'''
        languages = self.getAttribute(o, 'languages')
        if (len(languages) > 1) and value and \
           (not isinstance(value, dict) or (len(value) != len(languages))):
            raise Exception(INVALID_MULTILINGUAL_VALUE % self.name)
        Field.store(self, o, value)

    def storeValueFromAjax(self, o, value, currentValues):
        '''Stores p_value on p_o, taking care of multilinguality'''
        # Call the base method if p_self is not multilingual
        if not self.isMultilingual(o):
            return Field.storeValueFromAjax(self, o, value, currentValues)
        # Manage a multilingual field
        if currentValues is not None:
            # Take a copy of currentValues because it is mutable (dict)
            data = currentValues[self.name]
            if data is not None: data = data.copy()
            currentValues[self.name] = data
        # We get a partial value, for one language only
        language = o.req.languageOnly
        v = self.getUniStorableValue(o, value)
        o.values[self.name][language] = v
        return ' (%s)' % language
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Management of the model behind a Appy application'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import pathlib
from appy.model.workflow import Role

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Model configuration'''
    def __init__(self):
        # The "root" classes are those that will get their menu in the user
        # interface. Put their names in the list below. If you leave the list
        # empty, all classes will be considered root classes (the default). If
        # rootClasses is None, no class will be considered as root.
        self.rootClasses = []
        # People having one of these roles will be able to create instances
        # of classes defined in your application.
        self.defaultCreators = ['Manager']
        # Roles in use in a Appy application are identified at make time from
        # workflows or class attributes like "creators": it is not needed to
        # declare them somewhere. If you want roles that Appy will be unable to
        # detect, add them in the following list. Every role can be a Role
        # instance or a string.
        self.additionalRoles = []
        # What roles can unlock locked pages ? By default, only a Manager can do
        # it. You can place here roles expressed as strings or Role instances,
        # global or local.
        self.unlockers = ['Manager']
        # If the following attribute is True, a comment within an object's
        # history will be editable, either by a Manager or by the user having
        # performed the corresponding action.
        self.editHistoryComments = False

    def set(self, appFolder):
        '''Sets site-specific configuration elements'''
        # The absolute path to the app as a pathlib.Path instance
        self.appPath = pathlib.Path(appFolder)
        # The application name
        self.appName = self.appPath.name

    def get(self, config, logger=None, appOnly=False):
        '''Creates and returns a Model instance (see below)'''
        from appy.model.loader import Loader
        return Loader(self, config, logger, appOnly).run()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Model:
    '''Represents an application's model = the base Appy model, completed and
       extended by the application model.'''
    class Error(Exception): pass

    # Names of base classes forming the base Appy model, in package appy.model,
    # and standard workflows defined in package appy.model.workflows.standard.
    # These classes and workflows are injected into any app's model.
    baseClasses = ('Page', 'User', 'Group', 'Tool', 'Translation', 'Carousel',
                   'Document', 'Query')
    baseWorkflows = ('Anonymous', 'Authenticated', 'Owner', 'TooPermissive')

    def __init__(self, config, classes, workflows):
        '''The unique Model instance is created by the
           appy.model.loader.Loader.'''
        # The app's global config
        self.config = config
        # All Appy classes, keyed by their name
        self.classes = classes # ~{s_className: appy.model.meta.Class}~
        # All Appy worfklows, keyed by their name
        self.workflows = workflows # ~{s_className: appy.model.meta.Workflow}~
        # The global, grantable roles (will be computed later, at run-time only)
        self.grantableRoles = None

    def getClasses(self, type=None):
        '''Returns a list of classes (sorted by alphabetical order of their
           name) from self.classes. If p_type is:
           - None:       all classes are returned;
           - "class":    only Appy classes are iterated;
           - "workflow": only Appy workflows are iterated.
        '''
        if type is None:
            attributes = ('classes', 'workflows')
        else:
            attributes = ('classes',) if (type == 'class') else ('workflows',)
        # Build the result list
        r = []
        for name in attributes:
            r += list(getattr(self, name).values())
        r.sort(key=lambda k: k.name.lower())
        return r

    def getRootClasses(self):
        '''Returns the list of root classes for this app'''
        r = self.config.model.rootClasses
        if r is None: return () # No root class at all
        if not r:
            # We consider every "app" class as being a root class
            r = [c for c in self.getClasses(type='class') if c.type == 'app']
        else:
            r = [self.classes[name] for name in r]
        return r

    def getRoles(self, base=None, local=None, grantable=None):
        '''Produces a list of all the roles used within all workflows and
           classes defined in this app.'''
        # ----------------------------------------------------------------------
        # If p_base is...
        # ----------------------------------------------------------------------
        # True  | it keeps only standard Appy roles;
        # False | it keeps only roles which are specific to this app;
        # None  | it has no effect (so it keeps both roles).
        # ----------------------------------------------------------------------
        # If p_local is...
        # ----------------------------------------------------------------------
        # True  | it keeps only local roles. A local role is granted to a user
        #       | or group, is stored and applies on a single object only;
        # False | it keeps only global roles. A global role is granted to a
        #       | group or user and applies everywhere throughout the app,
        #       | independently of any object;
        # None  | it has no effect (so it keeps both roles).
        # ----------------------------------------------------------------------
        # If p_grantable is...
        # ----------------------------------------------------------------------
        # True  | it keeps only roles that a Manager can grant;
        # False | if keeps only ungrantable roles (ie those that are implicitly
        #       | granted by the system like role "Authenticated";
        # None  | it has no effect (so it keeps both roles).
        # ----------------------------------------------------------------------
        r = {} # ~{s_roleName: Role_role}~
        # Collect roles from workflow states and transitions
        for workflow in self.workflows.values():
            for name in workflow.attributes.keys():
                for elem in getattr(workflow, name).values():
                    for role in elem.getUsedRoles():
                        r[role.name] = role
        # Gather roles from "creators" attributes from every class
        for class_ in self.classes.values():
            creators = class_.getCreators()
            if not creators: continue
            for role in creators:
                r[role.name] = role
        # Get additional roles from the config
        for role in self.config.model.additionalRoles:
            if isinstance(role, str):
                role = Role(role)
            r[role.name] = role
        # Filter the result according to parameters and return a list
        return [role for role in r.values() if role.match(base,local,grantable)]

    def getGrantableRoles(self, o):
        '''Returns the list of global roles that can be granted to a user'''
        return [(role.name, o.translate('role_%s' % role.name)) \
                for role in self.grantableRoles]
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Pretty-prints chunks of Python code'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re, keyword

from appy.xml.escape import Escape
from appy.utils.client import Resource

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
NO_DATA  = 'No data as been returned.'
URL_KO   = 'Error getting %s. %s.'
PART_KO  = 'Invalid part specifier "%s".'
PT_KO    = 'Invalid part type "%s". Valid ones are: %s.'
METH_KO  = 'Wrong name "%s" for "%s". Expected <className>.<methodName>.'
NAME_KO  = '%s "%s" was not found.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PrettyPrinter:
    '''Transforms a chunk of Python code into a pretty-printed XHTML version'''

    # Pretty-printer-specific error class
    class Error(Exception): pass

    # Regular expressions and their components, for extracting parts of code
    # (classes, methods...)
    # ~~~
    # A group representing blanks at the start of the found class or method.
    # Counting blanks is important to find the place where the method or class
    # will end.
    blanks = '(?P<blanks> *)'
    blanksR = '(?P=blanks)' # A reference to this group
    # The regex part allowing to detect "detentation" at the end of the class or
    # method definition.
    dedent = '(?:^%s[\w#]|^[\w#]|\Z)' % blanksR
    # A template regular expression matching a class declaration
    classDecl = 'class +%s *(?:\(|\:).*?'
    # A template regular expression matching a complete method definition
    methodDef = '%s^%s(def +%%s *\(.*?)%s' % (classDecl, blanks, dedent)
    # A template regular expression matching a complete class definition
    classDef = '^%s(%s)%s' % (blanks, classDecl, dedent)

    # Types of code parts that can be used in a part specifier, with their
    # corresponding regular expressions.
    partTypes = {
      'c': classDef,  # The signature of a class (name and docstring)
      's': methodDef, # The signature of a method (name and docstring)
      'm': methodDef  # A complete method (signature and body)
    }
    # Names of part types, used in messages
    partNames = {'c': 'class', 's': 'method', 'm': 'method'}

    # States, while parsing a Python object (method or class) definition
    IN_DEF  = 0 # Within the object definition ("def m1(self):", "class C1:")
    IN_BODY = 1 # Within the method body (after the "def" or "class" line)
    IN_DOC  = 2 # Within the object's docstring

    # Docstring delimiters
    docDelimiters = ('"""', "'''")

    # Regular expression identifying, in Python comments, Appy names: p_param1,
    # m_method2...
    commentNames = re.compile('(?<=\W)(p|m|r|c)_(\w+)')

    # Styles to apply to names matched by regular expression "commentNames"
    commentStyles = {'p': 'u', 'm': 'i', 'r': 'ui', 'c': 'bi'}

    # Regular expression matching Python keywords
    pythonKeywords = re.compile('(?<=\W)(%s)(?![\w-])'%'|'.join(keyword.kwlist))
    
    def __init__(self, code, part, url=None, o=None):
        # The content of the Python file to pretty-print
        self.code = code
        # A specifier indicating which portion of the file must be kept
        self.partType, self.partName = self.getPartElements(part)
        # If the p_code file came from a given URL, p_url is present and the
        # source object whose Rich field referenced this URL, as p_o.
        self.url = url
        self.o = o

    def prettifyKeywords(self, match):
        '''Formats a Python keyword in bold'''
        return '<b>%s</b>' % match.group(1)

    def prettifyCommentName(self, match):
        '''Formats a comment name identified in this m_match'''
        # Get the HTML tags to use to surround the name
        tags = self.commentStyles.get(match.group(1))
        if not tags:
            # An unknown type of name
            return match.group(0)
        start = end = ''
        for tag in tags:
            start = '%s<%s>' % (start, tag)
            end = '</%s>%s' % (tag, end)
        return '%s%s%s' % (start, match.group(2), end)

    def prettifyLoc(self, r, loc, inComment=False, css=None):
        '''Compute p_loc's pretty version and add it to p_r'''
        ploc = Escape.xml(loc)
        if inComment:
            # Format Appy-specific comment names
            ploc = self.commentNames.sub(self.prettifyCommentName, ploc)
            if css: ploc = '<span class="%s">%s</span>' % (css, ploc)
        else:
            # Format Python keywords
            ploc = self.pythonKeywords.sub(self.prettifyKeywords, ploc)
        r.append(ploc)

    def prettify(self, odef, blanks, signOnly):
        '''Returns the prettified version of p_odef, which is a class or method
           definition. If p_signOnly is True, from p_odef, we only keep the
           object's signature, and starting comments.'''
        first = True
        r = []
        state = self.IN_DEF
        add = self.prettifyLoc
        for loc in odef.split('\n'): # loc = line of code
            if state == self.IN_DEF:
                # We are still in the object definition
                if first:
                    add(r, '%s%s' % (blanks, loc))
                    first = False
                else:
                    add(r, loc)
                if loc.rstrip().endswith(':'):
                    # The end of the definition has been reached
                    state = self.IN_BODY
            elif state == self.IN_BODY:
                # We are in the method body
                stripped = loc.strip()
                prefix = stripped[:3]
                if prefix == '# ~':
                    # A special comment indicating that, if p_signOnly mode, we
                    # have to stop now.
                    if signOnly:
                        break
                    else:
                        add(r, loc)
                elif not stripped or (stripped[0] == '#'):
                    # Add this blank or comment line into the result
                    add(r, loc, inComment=True, css='comment')
                elif prefix in self.docDelimiters:
                    # We enter into a docstring
                    add(r, loc, inComment=True, css='docstring')
                    if (len(stripped) > 3) and \
                       (stripped[-3:] in self.docDelimiters):
                        # We are already out of it
                        pass
                    else:
                        state = self.IN_DOC
                else:
                    # This is a not-commented line of code. Stop here if we are
                    # in p_signOnly mode.
                    if signOnly:
                        break
                    else:
                        add(r, loc)
            elif state == self.IN_DOC:
                add(r, loc, inComment=True, css='docstring')
                stripped = loc.strip()
                if stripped[-3:] in self.docDelimiters:
                    # We are out of the docstring
                    state = self.IN_BODY
        r = '\n'.join(r)
        if self.url and self.o:
            r = '<a href="%s" target="appyIFrame" ' \
                'onclick="openPopup(\'iframePopup\',null,750,600)"/>' \
                '<img src="%s" class="picto"/></a>%s' % \
                (self.url, self.o.buildUrl('page.svg'), r)
        return r

    def run(self):
        '''Returns, as a string, a chunk of XHTML code representing the
           pretty-printed version of some p_self.part of this p_self.code.'''
        # Create the regular expression allowing to search p_self.part
        template = self.partTypes[self.partType]
        if self.partNames[self.partType] == 'method':
            template = template % tuple(self.partName.split('.'))
        else:
            template = template % self.partName
        rex = re.compile(template, re.S|re.M)
        # Extracts the part of p_self.code corresponding to p_self.part
        match = rex.search(self.code)
        if match is None:
            # The part was not found, return an error message
            isError = True
            content = NAME_KO % (self.partNames[self.partType].capitalize(),
                                 self.partName)
        else:
            # The part was found, return it
            blanks = match.group(1) # Leading blanks preceding the definition
            content = match.group(2)
            isError = False
            # Keep only the object "signature" (=declaration + starting
            # comments) or keep the entire definition ?
            signOnly = self.partType != 'm'
            # Prettify the code
            content = self.prettify(content, blanks, signOnly=signOnly)
        return self.wrap(content, isError=isError, escape=isError)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Class methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @classmethod
    def wrap(class_, content, isError=True, escape=True):
        '''Wraps and xml-escape (if p_escape is True) the final pretty p_content
           (or error message if p_isError is True) in a "pre" tag.'''
        prefix = '<i>' if isError else ''
        suffix = '</i>' if isError else ''
        if escape: content = Escape.xml(content)
        return '<pre>%s%s%s</pre>' % (prefix, content, suffix)

    @classmethod
    def getPartElements(class_, part):
        '''If p_part specifier is valid, this method returns the p_part,
           splitted into its elemens: (type, name). Else, it raises an
           exception.'''
        # The part must contain a type and a name
        try:
            type, name = part.split('_', 1)
        except ValueError:
            raise class_.Error(PART_KO % part)
        # The type must be among valid ones
        if type not in class_.partTypes:
            validTypes = ', '.join(class_.partTypes.keys())
            raise class_.Error(PT_KO % (type, validTypes))
        # For types "s" and "m", name must be of the form
        #                  <className>.<methodName>
        if (class_.partNames[type] == 'method') and (name.count('.') != 1):
            raise class_.Error(METH_KO % (name, type))
        return type, name

    @classmethod
    def cache(class_, url, data, o):
        '''Cache file content p_data downloaded at this p_url on the cache
           accessible via p_o.'''
        if o is None: return
        mainCache = o.H().cache
        if mainCache._pretty_:
            mainCache._pretty_[url] = data
        else:
            mainCache._pretty_ = {url: data}

    @classmethod
    def getFromCache(class_, url, o):
        '''Several parts of the same file at this p_url may be requested. So
           once a file has been downloaded once, cache its content. Get the
           p_url content from the cache if present.'''
        if o is None: return
        cache = o.H().cache._pretty_
        if cache is None: return
        return cache.get(url)

    @classmethod
    def getFromUrl(class_, url, part, o=None):
        '''Retrieve the code file present at this p_url and return its pretty-
           printed p_part.'''
        wrap = class_.wrap
        # Check p_part specifier before performing the HTTP request. If it is
        # invalid, we save a useless request.
        try:
            class_.getPartElements(part)
        except class_.Error as err:
            return wrap(str(err))
        # Get the code file and returns it wrapped in a "pre" tag. Get the file
        # content from the cache if present.
        try:
            data = class_.getFromCache(url, o)
            if data is None:
                content = Resource(url).get()
                data = getattr(content, 'body', None)
                class_.cache(url, data, o)
            if not data:
                r = wrap(URL_KO % (url, NO_DATA))
            else:
                r = class_(data, part, url=url, o=o).run()
        except Resource.Error as err:
            r = wrap(URL_KO % (url, str(err)))
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>Table test.</p>
<p>
<table class="plain">
<tbody>
<tr>
<td style="width:auto !important">Table 1 <br /></td>
<td colspan="2">aaaaa<br /></td>
</tr>
<tr>
<td>zzz <br /></td>
<td>
  <table cellspacing="1">
    <tr>
      <td style="font-size: 75%">SubTableA</td>
      <td>SubTableB</td>
    </tr>
    <tr>
      <td>SubTableC</td>
      <td>SubTableD</td>
    </tr>
  </table>
</td>
<td><b>Hello</b> blabla<table><tr><td>SubTableOneRowOneColumn</td></tr></table></td>
</tr>
<tr>
<td><p>Within a <b>para</b>graph</p></td>
<td><b>Hello</b> non bold</td>
<td>Hello <b>bold</b> not bold</td>
</tr>
</tbody>
</table>
</p>
<p></p>
<table border="1" cellpadding="5" cellspacing="5" style="width:500px">
 <tbody>
  <tr>
   <td style="background-color:rgb(255, 204, 51); color:rgb(84, 194, 33);">Fond en RVB et texte vert</td>
  </tr>
  <tr>
   <td style="background-color:#eb9eb9; color: #205c90">
    <p>Dans un &lt;p&gt; fond en # et texte bleu</p>
   </td>
  </tr>
  <tr>
   <td style="background-color:green; color: grey;">Fond en HTML color et texte gris</td>
  </tr>
 </tbody>
</table>
<p></p>
<table cellspacing="0">
<tr>
<th style="font-size:10px">Température</th>
<th colspan="2"><span style="font-size:10px">Janvier</span></th>
<th colspan="2"><span style="font-size:10px">Février</span></th>
<th colspan="2"><span style="font-size:10px">Mars</span></th>
<th colspan="2"><span style="font-size:10px">Avril</span></th>
<th colspan="2"><span style="font-size:10px">Mai</span></th>
<th colspan="2"><span style="font-size:10px">Juin</span></th>
<th colspan="2"><span style="font-size:10px">Juillet</span></th>
<th colspan="2"><span style="font-size:10px">Août</span></th>
<th colspan="2"><span style="font-size:10px">Septembre</span></th>
<th colspan="2"><span style="font-size:10px">Octobre</span></th>
<th colspan="2"><span style="font-size:10px">Novembre</span></th>
<th colspan="2"><span style="font-size:10px">Décembre</span></th>
</tr>
<tr>
<td style="font-size:9px" rowspan="5">Les cinq mois les plus chauds</td>
<td><span style="font-size:9px">1975</span></td>
<td><span style="font-size:9px">6.3</span></td>
<td><span style="font-size:9px">1990</span></td>
<td><span style="font-size:9px">7.9</span></td>
<td><span style="font-size:9px">1991</span></td>
<td><span style="font-size:9px">9.5</span></td>
<td><span style="font-size:9px">1987</span></td>
<td><span style="font-size:9px">11.9</span></td>
<td><span style="font-size: 9px; color: red;">1947</span></td>
<td><span style="font-size:9px">15.8</span></td>
<td><span style="font-size:9px">1976</span></td>
<td><span style="font-size:9px">19.3</span></td>
<td><span style="font-size:9px">1994</span></td>
<td><span style="font-size:9px">21.8</span></td>
<td><span style="font-size:9px">1997</span></td>
<td><span style="font-size:9px">21.2</span></td>
<td><span style="font-size:9px">1949</span></td>
<td><span style="font-size:9px">17.7</span></td>
<td><span style="font-size:9px">1921</span></td>
<td><span style="font-size:9px">14.0</span></td>
<td><span style="font-size:9px">1994</span></td>
<td><span style="font-size:9px">10.4</span></td>
<td><span style="font-size:9px">1934</span></td>
<td><span style="font-size:9px">7.5</span></td>
</tr>
</table>
<table cellspacing="0" cellpadding="0" border="0" style="width:1024px">
 <tbody>
 </tbody>
</table>
<p></p>
<table border="0">
  <tr><td>A table</td><td>without</td></tr>
  <tr><td colspan="2" align="center">b o r d e r</td></tr>
</table>
<p></p>
<table border="1" style="border:1px solid black;" class="TableKWN">
    <tbody>
        <tr>
            <td>1</td>
        </tr>
        <tr>
            <td>2</td>
        </tr>
    </tbody>
</table>
<table border="undefined"><tr><td>No border</td></tr></table>
<p style="text-indent: -1.00cm">Negative text indent kept (outside a table)</p>
<table style="margin-right:auto">
  <tbody>
    <tr>
    </tr>
    <tr>
      <td>Article 1</td>
      <td><p style="text-indent: -1.59cm">Content</p></td>
    </tr>
  </tbody>
</table>
'''



'''Module defining database indexes'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# The class hierarchy defined in this module is inspired by sub-package
# "PluginIndexes" within the "Products.ZCatalog" product published by the Zope
# foundation at https://github.com/zopefoundation/Products.ZCatalog. The
# objective is to get similar indexes, but independently of Zope.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import persistent
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree
from BTrees.IIBTree import IITreeSet, intersection

from appy.database.operators import Operator

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
INVALID_VALUE = 'Index "%s" in catalog "%s": wrong %s "%s".'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Index(persistent.Persistent):
    '''Abstract base class for any index'''

    # Index-specific exception class
    class Error(Exception): pass

    # Values considered as empty, non-indexable values
    emptyValues = (None, [], ())

    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # An index is made of 2 main dicts: "byValue" stores objects keyed by their
    # index values (this is the "forward" index), while "byObject" stores index
    # values keyed by object ID (this the "reverse" index).
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Let's take an example. Suppose we want to index values for some
    # attribute on 3 objects. Here are the attribute values.
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Object   | Attribute value
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # obj1     | value1
    # obj2     | value2
    # obj3     | value1
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Here are the content for dicts "byValue" and "byObject".
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Dict     | Value
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # byValue  | { value1: [obj1, obj3], value2: [obj2] }
    # byObject | { obj1: value1, obj2: value2, obj3: value1 }
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # In m_init below, you will get more precisions about the data structures in
    # use. For example, object integer IDs are stored (from their attribute
    # o.iid), and not the objects themselves. For more info about object IIDs,
    # see appy/database/__init__.py).
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def __init__(self, field, catalog):
        '''Index constructor'''
        # The name of the index is the name of this p_field on some Appy class
        self.name = field.name
        # The catalog storing this index
        self.catalog = catalog
        self.init()

    def __repr__(self):
        '''p_self's string representation'''
        return '<index %s::%s>' % (self.catalog.name, self.name)

    def init(self):
        '''Initialisation (or cleaning) of the index main data structures'''

        # Dict "byValue" is a OOBTree of the form ~{indexValue: objectIds}~
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # indexValue   | Is an index value, produced from an object's attribute
        #              | value via m_getIndexValue on the corresponding field.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # objectIds    | Is a list of object IDs having this index value,
        #              | implemented as a IITreeSet.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self.byValue = OOBTree()

        # Dict "byObjects" is a IOBTree of the form ~{i_objectId: indexValue}~
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # objectId     | Is the object IID as stored in attribute object.iid
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # indexValue   | Is the value currently stored by the index for this
        #              | object.
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self.byObject = IOBTree()

    def getByValue(self, value, field):
        '''Returns the object IDs matching this p_value by performing a lookup
           in p_self.byValue.'''
        return self.byValue.get(self.toTerm(value, field), None)

    def getByRange(self, lo, hi, field):
        '''Returns the object IDs matching, within p_self.byValue, the range of
           values specified by p_lo(w) and p_hi(gh).'''
        if hi is None:
            r = self.byValue.values(self.toTerm(lo, field))
        else:
            r = self.byValue.values(self.toTerm(lo, field),
                                    self.toTerm(hi, field))
        return r

    def getByObject(self, o, field):
        '''Returns the value currently stored for p_o in p_self'''
        r = self.byObject.get(o.iid)
        if not r: return
        # Take care of converting the value from the internal to the standard
        # data format.
        return self.fromIndexed(r, field)

    def isMultiple(self, value, inIndex=False):
        '''Is p_value considered to be "multiple"? If yes, p_value will be
           considered as a sequence of several values; each of them must have a
           separate entry in p_self.byValue.'''
        # If p_inIndex is False, p_value is stored in the database, on some
        # object. If p_inIndex is True, p_value comes from within an index.
        type = tuple if inIndex else list
        # Indeed, multiple values stored in the database are lists, while
        # multiple values are stored in indexes are tuples.
        return isinstance(value, type)

    def getMultiple(self, value):
        '''Return a copy of this multiple p_value ready to be stored in
           p_self.byObject.'''
        return tuple(value)

    def valueEquals(self, value, current):
        '''Returns True if p_value and p_current are equal'''
        if not self.isMultiple(value):
            r = value == current
        else:
            # p_value may be a list, and p_current a p_tuple
            size = len(value)
            r = size == len(current)
            if r:
                i = 0
                while i < size:
                    if value[i] != current[i]:
                        r = False
                        break
                    i += 1
        return r

    def removeByValueEntry(self, value, id):
        '''Removes (non-multiple) p_value from p_self.byValue, for this p_id'''
        ids = self.byValue.get(value)
        if ids is None: return
        # Remove reference to p_id for this p_value
        try:
            ids.remove(id)
        except (ValueError, KeyError):
            pass
        # If no object uses p_value anymore, remove the whole entry
        if not ids:
            del(self.byValue[value])

    def removeEntry(self, id):
        '''Remove the currently indexed value for this p_id'''
        # Remove the entry in dict "byObject"
        value = self.byObject[id]
        del(self.byObject[id])
        # Remove the entry (or entries) in dict "byValue"
        if self.isMultiple(value, inIndex=True):
            # A multi-value
            for v in value:
                self.removeByValueEntry(v, id)
        else:
            self.removeByValueEntry(value, id)

    def addEntry(self, id, value, byValueOnly=False):
        '''Add, in this index, an entry with this p_id and p_value.

           If p_byValueOnly is True, it updates p_self.byValue but not
           p_self.byObject. It happens when m_addEntry is called recursively,
           for adding multiple entries corresponding to a multi-value. '''
        if self.isMultiple(value):
            # A multi-value. Get it as a ready-to-store value.
            value = self.getMultiple(value)
            # Add an entry in p_self.byObject
            self.byObject[id] = value
            # Add one entry for every single value in p_self.byValue
            for v in value:
                self.addEntry(id, v, byValueOnly=True)
        else:
            # A single value. Add it to p_self.byObject when relevant.
            if not byValueOnly:
                self.byObject[id] = value
            # Add it in p_self.byValue
            if value in self.byValue:
                self.byValue[value].insert(id)
            else:
                self.byValue[value] = IITreeSet((id,))

    def indexObject(self, o):
        '''Index object p_o. Returns True if the index has been changed
           regarding p_o, ie, entries have been added or removed.'''
        # Get the value to index
        value = o.getField(self.name).getIndexValue(o)
        id = o.iid
        if value in Index.emptyValues:
            # There is nothing to index for this object
            if id not in self.byObject: return
            # The object is yet indexed: remove the entry
            self.removeEntry(id)
        else:
            # We must index this value
            if id not in self.byObject:
                # No trace from this object in the index, add an entry
                self.addEntry(id, value)
            else:
                # The object is already indexed. Get the currently indexed value
                current = self.byObject[id]
                # Do nothing if the current value is the right one
                if self.valueEquals(value, current): return
                # Remove the current entry and replace it with the new one
                self.removeEntry(id)
                self.addEntry(id, value)
        return True

    def unindexObject(self, iid):
        '''Unindex object whose IID is p_iid. Returns True if an entry has
           actually been removed from the index.'''
        # Do nothing if there is no trace from this object in this index
        if iid not in self.byObject: return
        self.removeEntry(iid)
        return True

    def search(self, value, field, rs):
        '''Computes the objects matching p_value, updates the global result
           set p_rs and returns the updated result.'''
        isOperator = isinstance(value, Operator)
        try:
            if isOperator:
                # p_value is an operator on one or several values
                r, rsUpdated = value.apply(self, field, rs)
            else:
                # A single value
                r = self.byValue.get(self.toTerm(value, field), None)
                rsUpdated = False
            if r is None:
                # No match for this index = no match at all
                return r
            else:
                # Applying the operator may already have updated the global
                # resultset p_rs.
                return r if rsUpdated else intersection(r, rs)
        except TypeError:
            # p_value (or one of the value operators if p_value is an operator)
            # is not a valid value for this index.
            term = 'operator' if isOperator else 'index value'
            raise self.Error(INVALID_VALUE % \
                             (self.name, self.catalog.name, term, str(value)))

    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Conversion methods
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @classmethod
    def toIndexed(class_, value, field):
        '''Returns p_value's internal index representation'''
        return value # Is overridden by some child index classes

    @classmethod
    def fromIndexed(class_, value, field):
        '''Converts p_value, the internal representation of a value in this
           index, into the effective value.'''
        return value # Is overridden by some child index classes

    @classmethod
    def toTerm(class_, value, field):
        '''p_value is used as term for performing a search. This method converts
           it in such a way that it can be used as a valid search term for
           performing searches in the index.'''
        # Most of the time, p_value's internal representation in the index is a
        # suitable search term.
        return class_.toIndexed(value, field)

    @classmethod
    def toString(class_, o, value):
        '''Returns the string representation of this p_value'''
        if value is None: return
        return value if isinstance(value, str) else str(value)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.fields import Field
from appy.database.operators import in_

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Integer(Field):
    view = cell = Px('''
     <x>::field.getInlineEditableValue(o, value, layout, name=name)</x>
     <input type="hidden" if="masterCss"
            class=":masterCss" value=":value" name=":name" id=":name"/>''')

    edit = Px('''
     <input type="text" id=":name" name=":name" size=":field.width"
            maxlength=":field.maxChars" readonly=":field.isReadonly(o)"
            value=":field.getInputValue(inRequest, requestValue, value)"
            style=":'text-align: %s' % field.alignOnEdit"
            autocomplete=":field.autoComplete and 'on' or 'off'"/>

     <script if="hostLayout" var2="x=o.Lock.set(o, user, field=field)">:\
      'prepareForAjaxSave(%s,%s,%s,%s)' % \
       (q(name), q(o.iid), q(o.url), q(hostLayout))</script>''')

    search = Px('''
     <!-- From -->
     <label lfor=":widgetName">:_('search_from')</label> 
     <input type="text" name=":widgetName" maxlength=":field.maxChars"
            value=":field.sdefault[0]" size=":field.swidth"/>
     <!-- To -->
     <x var="toName='%s_to' % name">
      <label lfor=":toName">:_('search_to')</label> 
      <input type="text" name=":toName" maxlength=":field.maxChars"
             value=":field.sdefault[1]" size=":field.swidth"/>
     </x><br/>''')

    @classmethod
    def inRange(class_, o, range, value):
        '''Is p_value within p_range ? If no, returns a translated message'''
        # p_range must be expressed as a tuple (min, max) or an object having
        # attributes "min" and "max".
        if isinstance(range, tuple):
            min, max = range
        else:
            min = range.min
            max = range.max
        if (value < min) or (value > max):
            r = o.translate('range_ko', mapping={'min': min, 'max': max})
        else:
            r = True
        return r

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None,
      emptyIndexValue=None, searchable=False, readPermission='read',
      writePermission='write', width=5, height=None, maxChars=13, colspan=1,
      master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, sdefault=('',''),
      scolspan=1, swidth=None, sheight=None, persist=True, inlineEdit=False,
      view=None, cell=None, edit=None, xml=None, translations=None,
      readonly=False, alignOnEdit='left', autoComplete=True):
        # If attribute "readonly" is True (or stores a method returning True),
        # the rendered input field, on edit layouts, will have attribute
        # "readonly" set.
        self.readonly = readonly
        # On "edit", the alignment of the number encoded in the value can be
        # "left", "right" or "center".
        self.alignOnEdit = alignOnEdit
        # Activate or not browser's field auto-completion
        self.autoComplete = autoComplete
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          emptyIndexValue, searchable, None, readPermission, writePermission,
          width, height, maxChars, colspan, master, masterValue, focus,
          historized, mapping, generateLabel, label, sdefault, scolspan, swidth,
          sheight, persist, inlineEdit, view, cell, edit, xml, translations)
        # Define the corresponding Python type for values stored for this type
        self.pythonType = int
        # Define the filter PX when appropriate
        if self.indexed:
            self.filterPx = 'pxFilterText'

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        return '' if self.isEmptyValue(o, value) else str(value)

    def replaceSeparators(self, value):
        '''While integer values include no separator, sub-classes (like Float)
           may have it. This method replaces separators in such a way that
           p_value can be a valid Python literal value.'''
        return value

    def validateValue(self, o, value):
        # Replace separators when relevant
        value = self.replaceSeparators(value)
        try:
            value = self.pythonType(value)
        except ValueError:
            return o.translate('bad_%s' % self.pythonType.__name__)

    def isSortable(self, inRefs=False):
        '''Can this field be sortable ?'''
        return True if inRefs else Field.isSortable(self) # Sortable in Refs

    def getStorableValue(self, o, value, complete=False):
        if not self.isEmptyValue(o, value):
            value = self.replaceSeparators(value)
            return self.pythonType(value)

    def searchValueIsEmpty(self, req):
        '''Is there a search value or interval specified in p_form for this
           field ?'''
        # The base method determines if the "from" search field is empty
        isEmpty = Field.searchValueIsEmpty
        # We consider the search value being empty if both "from" and "to"
        # values are empty.
        return isEmpty(self, req) and \
               isEmpty(self, req, widgetName='%s_to' % self.name)

    def getSearchValue(self, req, value=None):
        '''Converts the raw search value from p_req into a single or an
           interval of typed values.'''
        # Get the "from" value
        value = Field.getSearchValue(self, req, value=value)
        index = self.getIndexType(True)
        if value:
            value = self.pythonType(value)
        # Get the "to" value
        toValue = Field.getSearchValue(self, req, '%s_to' % self.name)
        if toValue:
            toValue = self.pythonType(toValue)
        # Return an interval, excepted if both values are the same
        if value != toValue:
            r = in_(value, toValue)
        else:
            r = value
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
